//--------------------------------------------------------------------------//
// xdata8.cc                                                                //
// Latest revision: 05-23-2022                                              //
// Non linear curve fitting                                                 //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;
int status, singular, errpos, iteration;
const int MAXPARAMS=20;
double dyda[MAXPARAMS];           
double a2[MAXPARAMS];           
char *formula;
double beta[MAXPARAMS];           
double atry[MAXPARAMS];           
double da[MAXPARAMS];           
double minerr = 0.000001;  // Difference in chi sq to regard as converged 
char **equations=NULL;      


//--------------------------------------------------------------------------//
// marq                                                                     //
// Input requirements:                                                      //
//  ndata       = no.of data points                                         //
//  x(i,mx),x(i,my)   = data points                                         //
//  x(i,ms)           = individual std.deviations for data points           //
//  ma          = no.of coefficients                                        //
//  Funcs(xx,a,yfit,dyda,ma) = a subroutine [e.g.Fgauss] to evaluate the    //
//    fitting function yfit and its derivatives dyda with respect to        //
//    the fitting parameters  A at X.                                       //
//    On the first call to MRQMIN an initial guess is made for parame-      //
//    ters A, and alamda  is set <0 to initialize.  MRQMIN  is called       //
//    repeatedly until convergence is achieved.                             //
//    equation() = user-defined formula                                     //
//--------------------------------------------------------------------------//
void marq(char **arg, int noofargs)
{
  static int hit=0;
  static int funcID=0;           // which equation to fit 
  static int mx=1;               // column for x
  static int my=2;               // column for y
  static int ms=4;               // column for std dev
  static int mz=5;               // column for results
  static int mp=6;               // column for result parameters
  static int me=7;               // column for % errors
  static int mfit=2;             // no. of params to fit                
  static int usems=0;            // 1=user-supplied std err for data
  static double a[MAXPARAMS];    // parameters
  static int maxiterations = 200;// max. no. of iterations
  arg=arg; noofargs=noofargs;
  int j,k,index;
  int ma = MAXPARAMS;
  int nca = MAXPARAMS;
  int ndata=0;                  // no of data points
  double chisq=0.0;             // chi square of estimate
  double alamda;
  double **alpha, **covar, **da2;
  int lista[MAXPARAMS];           
  double sumx, sumy, ppx, ppy, meanx, meany, xxx, yyy, xvariance, yvariance, 
         sdx, sdy, see = 0.0, sex=0.0, sey=0.0, sb=0.0, sxsy, bb, cc;
//  double sb1, sb2;
  int nn = ndata;
  int i2 = ndata - mfit - 1;


  formula =  new char[1024];
  alpha = new double *[MAXPARAMS];
  covar = new double *[MAXPARAMS];
  da2 = new double *[MAXPARAMS];
  for(k=0;k<MAXPARAMS;k++) alpha[k] = new double[MAXPARAMS];
  for(k=0;k<MAXPARAMS;k++) covar[k] = new double[MAXPARAMS];
  for(k=0;k<MAXPARAMS;k++) da2[k] = new double[MAXPARAMS];
  if(!hit)
  {    for(k=0;k<MAXPARAMS;k++) a[k] = 1.0;
  }

  ////------------------------////
  //// Start of menu section  ////
  ////------------------------////

  if(equations==NULL)
  {   equations = new char*[100];
      for(k=0;k<100;k++) equations[k] = new char[128];
      strcpy(equations[0],"Polynomial        y=a0 + a1*x + a2*x^2 + ...   ");
      strcpy(equations[1],"Inv. Polynomial   y=a0 + a1*x^-1 + a2*x^-2+ ...");
      strcpy(equations[2],"Gaussian peaks (max=3 peaks, 3 parameters/peak)");
      strcpy(equations[3],"Exponentials      y=a0 + a1*exp(a2*x) + ...    ");
      strcpy(equations[4],"Saturation curve  y = a0*x/(a1+x)              ");
      strcpy(equations[5],"Inhibitor eq.     y = x/(a0*x^2 + a1)          ");
      for(k=6;k<100;k++) equations[k][0] = 0;
  }
  Dialog dialog;
  dialog.needrepeat = 0;  

  strcpy(dialog.title,"Non linear regression");
  strcpy(dialog.boxes[0],"Equation");
  strcpy(dialog.boxes[1],"Initial parameters");
  strcpy(dialog.boxes[2],"No. of parameters to fit");
  strcpy(dialog.boxes[3],"Max. no. of iterations");
  strcpy(dialog.boxes[4],"Min. chi square error");
  strcpy(dialog.boxes[5],"Data");
  strcpy(dialog.boxes[6],"Column for x   ");
  strcpy(dialog.boxes[7],"Column for y   ");
  strcpy(dialog.boxes[8],"Column for SD (check box if known)");
  dialog.boxset[8] = usems;
  strcpy(dialog.boxes[9],"Results");
  strcpy(dialog.boxes[10],"Column for estimated y's");
  strcpy(dialog.boxes[11],"Column for fitted params");
  strcpy(dialog.boxes[12],"Column for % errors");

  dialog.boxtype[0]=LIST;
  dialog.boxmin[0]=0;           // min menu selection
  dialog.boxmax[0]=100;         // max no.of menu items to use
  dialog.boxmenutitle[0] = new char[128];
  dialog.boxmenu[0] = equations;
  dialog.boxmenuselection[0] = funcID;
  strcpy(dialog.boxmenutitle[0],"Select or edit equation");
  for(j=0;j<DCOUNT;j++) dialog.startcol[j]=0;
  dialog.list_editrow[0] = 6;
  dialog.list_editcol[0] = 0; 
  dialog.list_maxstringsize[0] = 127; 

  dialog.boxtype[1]=LIST;
  dialog.boxmin[1]=0;           // min menu selection
  dialog.boxmax[1]=MAXPARAMS-1; // max no.of menu items to use
  dialog.boxmenutitle[1] = new char[128];
  dialog.boxmenu[1] = new char*[MAXPARAMS];
  for(k=0;k<MAXPARAMS;k++) dialog.boxmenu[1][k] = new char[64];
  dialog.boxmenuselection[1]=0;
  strcpy(dialog.boxmenutitle[1],"Initial parameters");
  for(j=0;j<MAXPARAMS;j++) 
      sprintf(dialog.boxmenu[1][j],"a%d= %g",j,a[j]);
  dialog.list_editrow[1] = 0;
  dialog.list_editcol[1] = 4; 
  dialog.list_maxstringsize[1] = 63; 

  dialog.boxtype[2]=INTCLICKBOX;
  sprintf(dialog.answer[2],"%d",mfit);
  dialog.boxmin[2]=1; 
  dialog.boxmax[2]=MAXPARAMS-1;

  dialog.boxtype[3]=INTCLICKBOX;
  sprintf(dialog.answer[3],"%d",maxiterations);
  dialog.boxmin[3]=1; 
  dialog.boxmax[3]=10000;

  dialog.boxtype[4]=STRING;
  sprintf(dialog.answer[4],"%g",minerr);

  dialog.boxtype[5]=LABEL;

  dialog.boxtype[6]=INTCLICKBOX;
  sprintf(dialog.answer[6],"%d",mx);
  dialog.boxmin[6]=1; 
  dialog.boxmax[6]=s.cols;

  dialog.boxtype[7]=INTCLICKBOX;
  sprintf(dialog.answer[7],"%d",my);
  dialog.boxmin[7]=1; 
  dialog.boxmax[7]=s.cols;

  dialog.boxtype[8]=TOGGLESTRING;
  sprintf(dialog.answer[8],"%d",ms);

  dialog.boxtype[9]=LABEL;

  dialog.boxtype[10]=INTCLICKBOX;
  sprintf(dialog.answer[10],"%d",mz);
  dialog.boxmin[10]=1; 
  dialog.boxmax[10]=s.cols;

  dialog.boxtype[11]=INTCLICKBOX;
  sprintf(dialog.answer[11],"%d",mp);
  dialog.boxmin[11]=1; 
  dialog.boxmax[11]=s.cols;

  dialog.boxtype[12]=INTCLICKBOX;
  sprintf(dialog.answer[12],"%d",me);
  dialog.boxmin[12]=1; 
  dialog.boxmax[12]=s.cols;

  for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
  for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
  for(j=0;j<20;j++) dialog.boxcomeback[j]=0;

  dialog.noofradios=0;
  dialog.noofboxes=13;
  dialog.helptopic=0;
  dialog.keep=0;
  dialog.needrepeat=0;
  dialog.want_changecicb = 0;
  dialog.f1 = null;
  dialog.transient = 1;
  strcpy(dialog.path, ".");

  ///////---///////
  dialogbox(dialog);
  ///////---///////

  if(g.getout){ g.getout=0; goto marqdone; }

  funcID = dialog.boxmenuselection[0];
  strcpy(formula, dialog.boxmenu[0][funcID]);

  ////  Equation
  if(funcID>5)        // chop off the "y="
  {   index = string_index(dialog.boxmenu[0][funcID],1,0,'=')+1;
      if(index>0)strcpy(formula, formula+index);
  }
  
  ////  Initial estimates (can't be 0)
  for(j=0;j<MAXPARAMS;j++) 
  {   index = string_index(dialog.boxmenu[1][j],1,0,'=')+1;
      a[j] = atof(dialog.boxmenu[1][j]+index);
      if(fabs(a[j])<1e-20) a[j] = 1e-20;
  }

  mfit = atoi(dialog.answer[2]);
  maxiterations = atoi(dialog.answer[3]);
  minerr = atof(dialog.answer[4]);
  mx = atoi(dialog.answer[6]);
  my = atoi(dialog.answer[7]);
  ms = atoi(dialog.answer[8]);
  mz = atoi(dialog.answer[10]);
  mp = atoi(dialog.answer[11]);
  me = atoi(dialog.answer[12]);
  usems = dialog.boxset[8];


  ////  End of menu section   ////
  ////------------------------////
  ////  Marquardt-Levenberg   ////

  for(k=0; k<mfit; k++) lista[k] = k;  // kth parameter
  ndata = s.maxrow[my]; 
  nca = mfit;
  iteration = 0;
  status = START;

  //// Don't look at this
  //// Change from base 1 to base 0
  for(j=0;j<ndata;j++)
  {
     s.x[mx][j]=s.x[mx][j+1];
     s.x[my][j]=s.x[my][j+1];
  }

  //// Std dev of each point is unknown so set to 1
  if(!usems)
      for(k=0; k<ndata; k++) s.x[ms][k] = 1;  // std. deviations for each data point
  
  alamda = -1;                            // 0 = initialize
  singular = 0; g.getout = 0; errpos = 0;

  //// Don't put a break in a while loop, it jumps out of the entire function
  while(status != DONE && !singular && !g.getout && !errpos)
  {   iteration++;
      mrqmin(funcID, my, mx, ms, ma, ndata, mfit, nca, lista, chisq, alamda, a, 
          covar, alpha, da2);
      if(iteration <= maxiterations) 
      {

         print_parameters(funcID, mfit, ndata, a, covar, alamda, chisq, usems);
         printf("\nIteration no. %d Chi sq.= %g\n",iteration, chisq);
         print_matrix((char*)"Covariance matrix:", covar, mfit, mfit);
         estimate_values(funcID, mx, my, mz, mp, me, ndata, a, dyda, mfit);
      }else
         status = DONE;
  }

  alamda = 0;
  mrqmin(funcID, my, mx, ms, ma, ndata, mfit, nca, lista, chisq, alamda, a, 
      covar, alpha, da2);
  print_parameters(funcID, mfit, ndata, a, covar, alamda, chisq, usems);
  status = OK;
  print_matrix((char*)"Covariance matrix:", covar, mfit, mfit);
  estimate_values(funcID, mx, my, mz, mp, me, ndata, a, dyda, mfit); 
  if(g.getout) message((char*)"Calculatus interruptus, ouch!");
  if(singular) message((char*)"It is singular!");

  //// stuff needed to compare two regression lines
  ////  x = data[0], y = data[1]

  nn = ndata;
  i2 = ndata - mfit - 1;


  sumx = 0.0; xvariance = 0.0;
  sumy = 0.0; yvariance = 0.0;
  ppx  = 0.0; 
  ppy  = 0.0;   
  for(j=0; j<nn; j++)
  {     sumx += s.x[mx][j]; 

        sumy += s.x[my][j]; 
  }
  meanx = sumx / (double)nn;
  meany = sumy / (double)nn;
  for(j=0; j<nn; j++) 
  { 
        xxx = s.x[mx][j] - meanx;          
        yyy = s.x[my][j] - meany;
        xvariance += ( ppx = xxx * xxx);
        yvariance += ( ppy = yyy * yyy);
  }
  xvariance /= (nn - 1);
  sdx = sqrt(xvariance);
  sdy = sqrt(yvariance);
  sxsy = sdx * sdy;
  sex = sdx / sqrt(nn);
  sey = sdy / sqrt(nn);
  sex = sex; //// shut up compiler
  sey = sey; //// shut up compiler
  see = sqrt(fabs(covar[1][1]));
  see *= sqrt(chisq/(double)(nn-2)); 
//  cc  = fabs(covar[1][1]) / sxsy;
//  cc = sxsy / 
  bb  = a[1];

  //// corr(x,y) = cov(x,y) / (sd of x * sd of y)
  //// cc        = cov(x,y) / sxsy

  printf("\nParameters needed for stats on two lines\n");
  printf("Use caution, only accurate for lines\n");
  sxsy = sdx * sdy;
  printf("sdx = %g\n",sdx);
  printf("sdy = %g\n",sdy);
  printf("sex = %g\n",sex);
  printf("sey = %g\n",sey);
  printf("sxsy = %g\n\n",sxsy);
  printf("covariance = %g\n",sxsy * a[1]);

  sb = see;
  printf("SB        = %g\n",sb);
  printf("B (slope) = %g\n",bb);
  printf("N         = %d\n",nn);

  cc = covar[1][0] * covar[0][1] / (covar[1][1] * covar[0][0]);
  printf("cc        = %g\n",cc);               // coef of correl

  i2 = i2; //// shut up compiler
//  for(jj=0; jj<mfit; jj++)
//  for(ii=0; ii<mfit; ii++)
//  {  cc = fabs(covar[ii][jj]);
//     see = sqrt(cc / (double)i2);                 // std.error of estimate
//     sb = see / (sdx * sqrt(nn - 1));
//     printf("x %d y %d cc = %g  sb = %g\n",ii, jj, cc,  sb);
//  }

//  cc = fabs(covar[0][0]);
//  see = sqrt(cc / (double)i2);                 // std.error of estimate
//  sb1 = see / (sdx * sqrt(nn - 1));
//  cc = fabs(covar[1][0]);
//  see = sqrt(cc / (double)i2);                 // std.error of estimate
//  sb2 = see / (sdx * sqrt(nn - 1));
//  cc = (sb2 / sb1) * (1.0 / sb) * a[1];
//  printf("cc        = %g\n",cc);               // coef of correl

marqdone:

  // Don't delete dialog.boxmenu[0]
  delete[] dialog.boxmenutitle[0];

  for(k=0;k<MAXPARAMS;k++) delete[] dialog.boxmenu[1][k];
  delete[] dialog.boxmenu[1];
  delete[] dialog.boxmenutitle[1];

  delete[] formula;
  for(k=0;k<MAXPARAMS;k++) delete[] da2[k];
  for(k=0;k<MAXPARAMS;k++) delete[] covar[k];
  for(k=0;k<MAXPARAMS;k++) delete[] alpha[k];
  delete[] da2;
  delete[] covar;
  delete[] alpha;
  hit=1;

  //// Don't look at this
  //// Change from base 1 to base 0
  for(j=ndata+1; j>0; j--)
  {
     s.x[mx][j] = s.x[mx][j-1]; // x values
     s.x[my][j] = s.x[my][j-1]; // y values
     s.x[mz][j] = s.x[mz][j-1]; // fitted y
     s.number[mx][j]=1;
     s.number[my][j]=1;
     s.number[mz][j]=1;
  }
  for(j=mfit; j>0; j--)
  {
     s.x[mp][j] = s.x[mp][j-1]; // params
     s.number[mp][j]=1;
  }
  s.maxrow[mp]++;
  refreshspreadsheet();
  XbaeMatrixRefresh(s.spreadsheet_widget);  

}


//--------------------------------------------------------------------------//
//  mrqmin                                                                  //
//--------------------------------------------------------------------------//
void mrqmin(int funcID, int my, int mx, int ms, int ma, int ndata, int mfit,  int nca,
     int *lista,  double &chisq, double &alamda,  double *a, double **covar, 
     double **alpha, double **da2)
{ 
   static double ochisq=1e12;
   static char tempstring[128];
   int ihit,j,k,k1,j1,kk;

   if(alamda < 0)
   {   kk = mfit;
       for(j=0; j<ma; j++)
       {
           ihit = 0;
           for(k=0; k<mfit; k++) if(lista[k] == j) ihit++;
           if(!ihit) lista[kk++] = j;
           else if(ihit > 1){ message((char*)"Improper permutation in LISTA"); return; }
       }
       if(kk != ma){ message((char*)"Improper permutation in LISTA"); return; }
       alamda = 0.001;
       mrqcof(funcID, mx, my, ms, ma, mfit, ndata, lista, a, alpha, beta, chisq);
       ochisq = chisq;
       if(chisq == 0){ status=DONE; message((char*)"Chi Square = 0 - ending calculations early"); return; }
       for(j=0; j<ma; j++) atry[j] = a[j];
   }
   for(j=0; j<mfit; j++)
   {   for(k=0; k<mfit; k++) covar[j][k] = alpha[j][k];
       covar[j][j] = alpha[j][j] * (1.0 + alamda);
       da[j] = beta[j];
   }
   for(j1=0; j1<mfit; j1++)
   for(k1=0; k1<mfit; k1++)
       da2[j1][k1] = da[j1];
   gaussj(covar, mfit, nca, da2, 1, 1); // matrix solution
   if(singular){ status=DONE; message((char*)"Singular matrix"); }
   for(k1=0; k1<mfit; k1++) da[k1] = da2[k1][0];
   printf("L= %g\n",alamda);
   if(alamda == 0)
   {    cosvrt(covar, ma, lista, mfit);
       status = DONE;
       return;
   }
   for(j=0; j<mfit; j++) atry[lista[j]] = a[lista[j]] + da[j];
   mrqcof(funcID, mx, my, ms, ma, mfit, ndata, lista, atry, covar, da, chisq);
   if(ochisq == 0 || chisq == 0){ status=DONE; message((char*)"Chi Square = 0 - ending calculations early"); return; }
   else if(fabs((chisq - ochisq)/ochisq) < minerr) 
   {   sprintf(tempstring,"Converged after\n%d iterations\n(Chi sq=%g)",iteration,chisq);
       message(tempstring); 
       status = DONE; 
   }
   if(chisq < ochisq)
   {   alamda = 0.1 * alamda;
       ochisq = chisq;
       for(j=0; j<mfit; j++) 
       {  for(k=0; k<mfit; k++) alpha[j][k] = covar[j][k];
          beta[j] = da[j];
          a[lista[j]] = atry[lista[j]];
       }
   }else if(chisq > ochisq)
   {   alamda *= 10;
       chisq = ochisq;
       if(alamda >= 1e+09){ message((char*)"Did not converge"); status = DONE; }
       if(chisq == 0){ status=DONE; message((char*)"Chi Square = 0 - Unable to calculate"); return; }
   }
}


//--------------------------------------------------------------------------//
//  gaussj                                                                  //
//--------------------------------------------------------------------------//
void gaussj(double **aa, int n, int np, double **bb, int m, int mp)
{
   // Var.  Inputs                                     Outputs
   // ----  -------------                              -----------------
   // aa()  n x np input matrix                        Inverse matrix
   // bb()  m x mp  matrix of right-hand side vectors  Solution vectors
  
  np=np;  mp=mp;
  double big, dum, pivinv;
  int ipiv[MAXPARAMS];           
  int indxr[MAXPARAMS];           
  int indxc[MAXPARAMS];           
  int i, j, k, L, LL, irow=0, icol=0;
  for(j=0; j<n; j++) ipiv[j]=0;
  for(i=0; i<n; i++)
  {   big = 0.0;
      for(j=0; j<n; j++)
      {   if(ipiv[j] != 1)
          {
             for(k=0; k<n; k++)
             {  if(ipiv[k] == 0)
                {   if(fabs(aa[j][k]) >= big) 
                    {  big = fabs(aa[j][k]);
                       irow = j;
                       icol = k;
                    }
                }else if(ipiv[k] > 1)
                {   message((char*)"Singular matrix");
                    status = DONE;
                    singular = 1;
                    return;
                }
             }
          }
      }
      ipiv[icol]++;
      if(singular) return;
      if(irow != icol)
      {   for(L=0;L<n; L++) dswap(aa[irow][L], aa[icol][L]);
          for(L=0;L<m; L++) dswap(bb[irow][L], bb[icol][L]);
      }
      indxr[i] = irow;
      indxc[i] = icol;
      if(aa[icol][icol] == 0)
      {    message((char*)"Singular matrix"); singular=1; status=DONE; return; }
      pivinv = divide(1.0, aa[icol][icol]);
      aa[icol][icol] = 1.0;
      for(L=0;L<n; L++) aa[icol][L] *= pivinv;
      for(L=0;L<m; L++) bb[icol][L] *= pivinv;
      for(LL=0;LL<n; LL++) 
      {   if(LL != icol)
          {   dum = aa[LL][icol];
              aa[LL][icol] = 0.0;
              for(L=0;L<n; L++) aa[LL][L] -= aa[icol][L] * dum;
              for(L=0;L<m; L++) bb[LL][L] -= bb[icol][L] * dum;
          }
      }
  }

  for(L=n-1;L>=0; L--) 
     if(indxr[L] != indxc[L])
         for(k=0;k<n; k++) dswap(aa[k][indxr[L]], aa[k][indxc[L]]); 
     
}


//--------------------------------------------------------------------------//
// divide                                                                   //
//--------------------------------------------------------------------------//
double divide(double a, double b)
{
   if(fabs(b) > 1e-200) return a/b;
   else return fsgn(a) * 1e100;
}


//--------------------------------------------------------------------------//
// mrqcof                                                                   //
//--------------------------------------------------------------------------//
void mrqcof(int funcID, int mx, int my, int ms, int ma, int mfit, int ndata, 
            int *lista, double *a, double **alpha, double *beta, double &chisq)
{
   ma=ma;
   double dy, sig2i, yy, wt;
   char tempstring[1024];
   int i,j,k;
   for(j=0; j<mfit; j++)
   {   for(k=0; k<=j; k++) alpha[j][k] = 0;
       beta[j] = 0;
   }
   chisq = 0.0;
   for(i=0; i<ndata; i++)
   {   yy = func(funcID, s.x[mx][i], a, dyda, mfit);
       if(s.x[ms][i] == 0.0)
       {  
           sprintf(tempstring, "SD of data point=0\nCheck column %d row %d\nor turn off SD data set\nUnable to calculate",ms,i);
           message(tempstring);
           status = DONE;
           break;
       }
       sig2i = divide(1, s.x[ms][i] * s.x[ms][i]);
       dy = (double)s.x[my][i] - yy;
       for(j=0; j<mfit; j++) 
       {   wt = dyda[lista[j]] * sig2i;
           for(k=0; k<=j; k++) alpha[j][k] += wt * dyda[lista[k]];
           beta[j] += dy * wt;
       }
       chisq += dy * dy * sig2i;
   }
   if(status == DONE) return;
   for(j=1; j<mfit; j++)
   for(k=0; k<j; k++)
      alpha[k][j] = alpha[j][k];
}



//--------------------------------------------------------------------------//
// cosvrt                                                                   //
//--------------------------------------------------------------------------//
void cosvrt(double **covar, int ma, int *lista, int mfit)
{
   int i,j;
   double swapem;
   for(j=0; j<ma-1; j++)
   for(i=j+1; i<ma; i++)
      covar[i][j] = 0.0;
   for(i=0; i<mfit-1; i++)
   for(j=i+1; j<mfit; j++)
   {  if(lista[j] > lista[i])
         covar[lista[j]][lista[i]] = covar[i][j];
      else
         covar[lista[i]][lista[j]] = covar[i][j];
   }
   swapem = covar[0][0];
   for(j=0; j<ma; j++)
   {  covar[0][j] = covar[j][j];
      covar[j][j] = 0.0;
   }
   covar[lista[0]][lista[0]] = swapem;
   for(j=0; j<mfit; j++)
      covar[lista[j]][lista[j]] = covar[0][j];
   for(j=1; j<ma; j++)
   for(i=0; i<=j-1; i++)
      covar[i][j] = covar[j][i];
}



//--------------------------------------------------------------------------//
// func = evaluates an equation                                             //
// xx   = x value                                                           //
// a    = array of parameters                                               //
// dyda = deriv of function, can be 1                                       //
// mfit = no of params                                                      //
//--------------------------------------------------------------------------//
double func(int funcID, double xx, double *a, double *dyda, int mfit)
{
    switch(funcID)
    { 
      case 0: return polynomial(xx, a, dyda, mfit);
      case 1: return poly2(xx, a, dyda, mfit); 
      case 2: return fgauss(xx, a, dyda, mfit);
      case 3: return mexp(xx, a, dyda, mfit); 
      case 4: return poly6(xx, a, dyda, mfit);
      case 5: return poly7(xx, a, dyda, mfit);
      default: return other(xx, a, dyda, mfit);
   }
}


//--------------------------------------------------------------------------//
// fgauss                                                                   //
//--------------------------------------------------------------------------//
double fgauss(double xx, double *a, double *dyda, int mfit)
{
  double yy = 0.0;
  int i;
  double arg=0, arg1, xex, fac; 

  //  use a(1 to ma) for the fitted parameters
  //  x(k,mx) and x(k,my) for data points
  //  dyda(1 to ma) for partial derivative of y with respect to each parameter
  //  name$(1 to ma) for names of parameters
  //  insert values for y=f(A(),xx),  name$(), and dyda()

  for(i=0;i<mfit;i+=3)
  {   if(a[i+2] == 0) a[i+2] = 0.0001;     
      arg = (xx - a[i + 1]) / a[i + 2];
      arg1 = arg;
      if(fabs(arg1) >= 7) arg1 = 7 * sgn(arg1);
      xex = exp(-arg1 * arg1);
      fac = a[i] * xex * 2 * arg;
      yy = yy + a[i] * xex;
      dyda[i] = xex;
      if(fabs(dyda[i]) < 1e-12) dyda[i] = 0.0;
      if(a[i+2] == 0) status = DONE;
      else
      {  dyda[i+1] = fac / a[i+2];
         dyda[i+2] = fac * arg / a[i+2];
      }
      if(fabs(dyda[i+1] < 1e-12)) dyda[i+1] = 0.0l; // prevent underflow when dyda is squared
      if(fabs(dyda[i+2] < 1e-12)) dyda[i+2] = 0.00;
  }
  return yy;
}



//--------------------------------------------------------------------------//
//  mexp                                                                    //
//--------------------------------------------------------------------------//
double mexp(double x, double *a, double *dyda, double mfit)
{
  // use a(1 to mfit) for the fitted parameters
  // x(k,mx) and x(k,my) for data points
  // dyda(1 to mfit) for partial derivative of y with respect to each parameter
  // name$(1 to mfit) for names of parameters
  // insert values for y=f(A(),x),  name$(), and dyda()

  double y;
  mfit=mfit; 
  y = a[0] + a[1]*exp(a[2]*x);
  if(mfit>3) y += a[3]*exp(a[4]*x);
  if(mfit>5) y += a[5]*exp(a[6]*x);
  if(mfit>7) y += a[7]*exp(a[8]*x);
  if(mfit>9) y += a[9]*exp(a[10]*x);
  if(mfit>11) y += a[11]*exp(a[12]*x);
  dyda[0] = 1;
  dyda[1] = exp(a[2]*x);
  dyda[2] = a[1]*x*exp(a[2]*x);
  if(mfit>3) dyda[3] = exp(a[4]*x);
  if(mfit>4) dyda[4] = a[3]*x*exp(a[4]*x);
  if(mfit>5) dyda[5] = exp(a[6]*x);
  if(mfit>6) dyda[6] = a[5]*x*exp(a[6]*x);
  if(mfit>7) dyda[7] = exp(a[8]*x);
  if(mfit>8) dyda[8] = a[7]*x*exp(a[8]*x);
  if(mfit>9) dyda[9] = exp(a[10]*x);
  if(mfit>10) dyda[10] = a[9]*x*exp(a[10]*x);
  if(mfit>11) dyda[11] = exp(a[12]*x);
  if(mfit>12) dyda[12] = a[11]*x*exp(a[12]*x);
  return y;
}



//--------------------------------------------------------------------------//
// other                                                                    //
//user-defined formula  using differentiation by 2-point subtraction        //
//  mfit = the no. of fitted parameters                                     //
//  a()  = the fitted parameters                                            //
//  dyda() =  partial derivative of y with respect to each parameter        //
//  name$() = names of parameters                                           //
//Insert values for y=f(A(),x),  and name$()                                //
//Substitute the name of the new function at 3 positions                    //
//Add a new function to calculate y, similar to poly2fnc                    //
//--------------------------------------------------------------------------//
double other(double x, double *a, double *dyda, int mfit)
{
   static char tempstring[128];
   int k;
   double y, y1=0.0;
   int orr;
   orr = g.rr;
   g.rr = cint(x);
   g.getout=0;



   ////  Set a[] parameters to interpreter
   for(k=0;k<mfit;k++)
   {   sprintf(tempstring, "a%d=%g", k, a[k]);
       eval(tempstring);
       if(g.getout) break;
   }


   ////  Calculate a value for y
   y = eval(formula);
   if(g.getout) return 0;
  
   ////  Calculate dyda's by difference. If it doesn't converge, change
   ////  the factor for a[k] in 2 places (they must total to 1.0).

   for(k=0;k<mfit;k++) a2[k] = a[k];
   for(k=0;k<mfit;k++)
   {   a2[k] = 0.9 * a[k];   // change 1 parameter
       sprintf(tempstring, "a%d=%f", k, a2[k]); 
       eval(tempstring);
       y1 = eval(formula);
       dyda[k] = divide(y - y1, .1 * a[k]);       // deriv.of y with resp.to a
       a2[k] = a[k];
       sprintf(tempstring, "a%d=%f", k, a2[k]); 
       eval(tempstring);
       if(g.getout) break;
   }
   g.rr = orr;
   return y;
}


//--------------------------------------------------------------------------//
//  estimate_values                                                         //
//--------------------------------------------------------------------------//
void estimate_values(int funcID, int mx, int my, int mz, int mp, int me, 
     int ndata, double *a, double *dyda, int mfit)
{
  int k;
  double y=0.0;
  //printf(" Obs. x       Obs. y         Est. y         %c Error\n", '%');
  //printf(" ------       ------         ------         -------\n");
  for(k=0;k<ndata;k++)
  {  y = func(funcID, s.x[mx][k], a, dyda, mfit);
     s.x[mz][k] = y;
     s.number[mz][k] = 1;
     s.string[mz][k] = 0;
     s.hex[mz][k] = 0;
     if(my>=0 && me>=0) s.x[me][k] = 100 * divide(s.x[my][k] - y,  s.x[my][k]);
     if(me>=0){ s.number[me][k] = 1;  s.string[me][k] = 1; }
  }
  for(k=0;k<mfit;k++)
  {  s.number[mp][k] = 1;
     s.string[mp][k] = 0;
     s.hex[mp][k] = 0;
     s.x[mp][k] = a[k];
  }
  s.maxrow[mz] = max(s.maxrow[mx], ndata);
  s.maxrow[mp] = max(s.maxrow[mp], mfit);
  if(me>=0) s.maxrow[me] = max(s.maxrow[me], ndata);
  s.maxcol = max(s.maxcol, mz);
  s.maxcol = max(s.maxcol, mp);
  if(me>=0) s.maxcol = max(s.maxcol, me);
  refreshspreadsheet();
}


//--------------------------------------------------------------------------//
//  erf                                                                     //
//--------------------------------------------------------------------------//
double erf(double x)
{
   double s1,s2,s3,s4,s5,y;
   s1 = fabs(x);
   s2 = 1 / (1 + .5 * s1);
   s3 = -.82215223 + s2 * .17087277;
   s3 = 1.48851587 + s2 * s3;
   s3 = -1.13520398 + s2 * s3;
   s3 = .27886807 + s2 * s3;            // Chebyshev fitting for
   s3 = -.18628806 + s2 * s3;           // estimation of
   s3 = s2 * (.09678418 + s2 * s3);     // complementary error function
   s3 = s2 * (1.00002368 + s2 * (.37049196 + s3));
   s5 = (-s1 * s1 - 1.26551223 + s3);
   if(s5 < -77) s5 = -77;
   s4 = s2 * exp(s5);
   if(x < 0) s4 = 2 - s4;
   y = s4;
   return y;
}



//--------------------------------------------------------------------------//
//  polynomial                                                              //
//--------------------------------------------------------------------------//
double polynomial(double x, double *a, double *dyda, int mfit)
{
   int k;
   double y = 0.0,y2;
   y = polynomial_evaluate(mfit,x,a); 
   for(k=0;k<mfit;k++) a2[k] = a[k];
   for(k=0;k<mfit;k++) 
   {  a2[k] = .99 * a[k];
      y2 = polynomial_evaluate(mfit, x, a2); 
      dyda[k] = divide(y - y2, 0.01 * a[k]);
      a2[k] = a[k];
   }
   return y;
}

//--------------------------------------------------------------------------//
//  polynomial_evaluate                                                     //
//--------------------------------------------------------------------------//
double polynomial_evaluate(int mfit, double x, double *a)
{
   x=x;
   int k;
   double answer = 0;
   for(k=0;k<mfit;k++) answer += a[k]*pow(x,k);
   return answer;
}


//--------------------------------------------------------------------------//
// poly2                                                                    //
// Polynomial estimation using differentiation by 2-point subtraction       //
//--------------------------------------------------------------------------//
double poly2(double x, double *a, double *dyda, int mfit)
{
   int k;
   double y = 0.0, y1=0.0, y2=0.0;
   y = poly2fnc(mfit, x, a);         // Substitute new function name here
   for(k=0;k<mfit;k++) a2[k] = a[k];
   y1 = poly2fnc(mfit, x, a2);       // Substitute new function name here
   for(k=0;k<mfit;k++) 
   {   a2[k] = .99 * a[k];
       y2 = poly2fnc(mfit, x, a2);   // Substitute new function name here
       dyda[k] = divide(y1 - y2, 0.01 * a[k]);
       a2[k] = a[k];
   }
   return y;
}



//--------------------------------------------------------------------------//
//  poly2fnc                                                                //
//--------------------------------------------------------------------------//
double poly2fnc(int mfit, double x, double *a)
{
   x=x;
   int k;
   double answer = 0;
   for(k=0;k<mfit;k++) answer += a[k]/pow(x,k);
   return answer;
}


//--------------------------------------------------------------------------//
// poly6                                                                    //
// Polynomial estimation using differentiation by 2-point subtraction       //
//--------------------------------------------------------------------------//
double poly6(double x, double *a, double *dyda, int mfit)
{
   int k;
   double y, y1, y2;
   y = 0.0;
   y = poly6fnc(mfit, x, a);         // Substitute new function name here
   for(k=0;k<mfit;k++) a2[k] = a[k];
   y1 = poly6fnc(mfit, x, a2);       // Substitute new function name here
   for(k=0;k<mfit;k++) 
   {   a2[k] = 0.99 * a[k];
       y2 = poly6fnc(mfit, x, a2);   // Substitute new function name here
       dyda[k] = divide(y1 - y2, 0.01 * a[k]);
       a2[k] = a[k];
   }
   return y;
}


//--------------------------------------------------------------------------//
//  poly6fnc                                                                //
//--------------------------------------------------------------------------//
double poly6fnc(int mfit, double x, double *a)
{
   mfit=mfit;
   return divide(a[0] * x, a[1] + x);
}


//--------------------------------------------------------------------------//
// poly7                                                                    //
// Polynomial estimation using differentiation by 2-point subtraction       //
//--------------------------------------------------------------------------//
double poly7(double x, double *a, double *dyda, int mfit)
{
   int k;
   double y = 0.0, y1, y2;
   y = poly7fnc(mfit, x, a);         // Substitute new function name here
   for(k=0;k<mfit;k++) a2[k] = a[k];
   y1 = poly7fnc(mfit, x, a2);       // Substitute new function name here
   for(k=0;k<mfit;k++) 
   {   a2[k] = .99 * a[k];
       y2 = poly7fnc(mfit, x, a2);   // Substitute new function name here
       dyda[k] = divide(y1 - y2, 0.01 * a[k]);
       a2[k] = a[k];
   }
   return y;
}


//--------------------------------------------------------------------------//
//  poly7fnc                                                                //
//--------------------------------------------------------------------------//
double poly7fnc(int mfit, double x, double *a)
{
   mfit=mfit;
   return  x / pow(a[0]*x, 2+a[1]);
}



//--------------------------------------------------------------------------//
//  print_parameters                                                        //
//--------------------------------------------------------------------------//
void print_parameters(int funcID, int mfit, int ndata, double *a, double **covar, 
    double alamda, double chisq, int usems)
{ 
   const int SPACING=14;
   int k, df;
   double sd=0.0, deltachisq1=0.0, deltachisq2=0.0, conlimit1, conlimit2;
   char show[256];
   char tempstring[256];
   char tempstring2[256];
   int len;
   if(funcID >= 8) sprintf(show, "y=%s", formula); else strcpy(show, formula);
   printf("Parameter Table     %s\n", show);
   printf("Param  Name    Value         1SD         68%c ConLimit 95%c ConLimit\n",'%','%');
   printf("-----  ------- ------        -----         -----------   ------------\n");
   df = ndata - mfit;
   deltachisq2 = 3.282 + 1.594 * df;   //  this formula is for 95% conf.level [2 sd]
   deltachisq1 = -.136 + 1.2046 * df;  //  this formula is for 1 sd
   for(k=0; k<mfit; k++)
   {  
        sprintf(tempstring, "  %d    a%d      ",k+1,k);
        sd = sqrt(fabs(covar[k][k]));

        //// If SD of each data point is unknown, assume good fit and
        //// calculate sd of estimate from chisq.
        //// SD of each data point is set to 1 above. 

        if(ndata>2 && usems==0) sd *= sqrt(chisq/(ndata-2));   

        conlimit1 = sqrt(fabs(deltachisq1)) * sd;
        conlimit2 = sqrt(fabs(deltachisq2)) * sd;

        len=strlen(tempstring);

        sprintf(tempstring2, "%g                  ",a[k]);
        tempstring[len]=0;
        strcat(tempstring, tempstring2);
        len+=SPACING;

        sprintf(tempstring2, "%g                  ",sd);
        tempstring[len]=0;
        strcat(tempstring, tempstring2);
        len+=SPACING;

        sprintf(tempstring2, "%g                  ",conlimit1);
        tempstring[len]=0;
        strcat(tempstring, tempstring2);
        len+=SPACING;

        sprintf(tempstring2, "%g                  ",conlimit2);
        tempstring[len]=0;
        strcat(tempstring, tempstring2);
        len+=SPACING;
        printf("%s\n",tempstring);

   }
}



//--------------------------------------------------------------------------//
//  print_matrix                                                            //
//--------------------------------------------------------------------------//
void print_matrix(char *title, double **mat, int x, int y)
{
  int i,j;
  printf("%s\n",title);
  for(j=0;j<x;j++)
  {   for(i=0;i<x;i++) printf("   %g ", mat[j][i]);
      printf("\n");
  }
}


//--------------------------------------------------------------------------//
//  evaluate                                                                //
//--------------------------------------------------------------------------//
void evaluate(char **arg, int noofargs)
{
  arg=arg; noofargs=noofargs;
  static int funcID=0;           // which equation to fit 
  static int mx=1;               // column for source x
  static int mp=2;               // column for source parameters
  static int mz=3;               // column for result y 
  static int mfit=2;             // no. of params to use
  int jj,kk,row;
  double value, xx;
  double dyda[20];               // deriv, set to 1
  double aa[20];                 // params
  char tempstring[1024];         // a module level global

  //// Start of menu section  ////

  ////  'equations' remains allocated forever
  if(equations==NULL)
  {   equations = new char*[100];
      for(kk=0; kk<100; kk++) equations[kk] = new char[128];
      strcpy(equations[0],"Polynomial        y=a0 + a1*x + a2*x^2 + ...   ");
      strcpy(equations[1],"Inv. Polynomial   y=a0 + a1*x^-1 + a2*x^-2+ ...");
      strcpy(equations[2],"Gaussian peaks (max=3 peaks, 3 parameters/peak)");
      strcpy(equations[3],"Exponentials      y=a0 + a1*exp(a2*x) + ...    ");
      strcpy(equations[4],"Saturation curve  y = a0*x/(a1+x)              ");
      strcpy(equations[5],"Inhibitor eq.     y = x/(a0*x^2 + a1)          ");
      for(kk=6; kk<100; kk++) equations[kk][0] = 0;
  }
  Dialog dialog;
  dialog.needrepeat = 0;  

  strcpy(dialog.title,"Evaluate equation");
  strcpy(dialog.boxes[0],"Equation");
  strcpy(dialog.boxes[1],"No. of parameters to fit");
  strcpy(dialog.boxes[2],"Data");
  strcpy(dialog.boxes[3],"Column for x   ");
  strcpy(dialog.boxes[4],"Column for params");
  strcpy(dialog.boxes[5],"Column for results");

  dialog.boxtype[0]=LIST;
  dialog.boxmin[0]=0;           // min menu selection
  dialog.boxmax[0]=100;         // max no.of menu items to use
  dialog.boxmenutitle[0] = new char[128];
  dialog.boxmenu[0] = equations;
  dialog.boxmenuselection[0] = funcID;
  strcpy(dialog.boxmenutitle[0],"Select or edit equation");
  for(jj=0; jj<DCOUNT; jj++) dialog.startcol[jj]=0;
  dialog.list_editrow[0] = 6;
  dialog.list_editcol[0] = 0; 
  dialog.list_maxstringsize[0] = 127; 

  dialog.boxtype[1]=INTCLICKBOX;
  sprintf(dialog.answer[1],"%d",mfit);
  dialog.boxmin[1]=1; 
  dialog.boxmax[1]=MAXPARAMS-1;

  dialog.boxtype[2]=LABEL;

  dialog.boxtype[3]=INTCLICKBOX;
  sprintf(dialog.answer[3],"%d",mx);
  dialog.boxmin[3]=1; 
  dialog.boxmax[3]=s.cols;

  dialog.boxtype[4]=INTCLICKBOX;
  sprintf(dialog.answer[4],"%d",mp);
  dialog.boxmin[4]=1; 
  dialog.boxmax[4]=s.cols;

  dialog.boxtype[5]=INTCLICKBOX;
  sprintf(dialog.answer[5],"%d",mz);
  dialog.boxmin[5]=1; 
  dialog.boxmax[5]=s.cols;

  for(jj=0; jj<10; jj++) for(kk=0; kk<20; kk++) dialog.radiotype[jj][kk]=RADIO;
  for(jj=0; jj<10; jj++) dialog.radiocomeback[jj]=0;
  for(jj=0; jj<20; jj++) dialog.boxcomeback[jj]=0;

  dialog.noofradios=0;
  dialog.noofboxes=6;
  dialog.helptopic=0;
  dialog.keep=0;
  dialog.needrepeat=0;
  dialog.want_changecicb = 0;
  dialog.f1 = null;
  dialog.transient = 1;
  strcpy(dialog.path, ".");

  ///////---///////
  dialogbox(dialog);
  ///////---///////

  if(g.getout){ g.getout=0; delete[] dialog.boxmenutitle[0]; return; }
  funcID = dialog.boxmenuselection[0];
  mfit = atoi(dialog.answer[1]); //// no of params
  mx = atoi(dialog.answer[3]);   //// x column
  mp = atoi(dialog.answer[4]);   //// params column
  mz = atoi(dialog.answer[5]);   //// results column

  ////------------------------////
  ////  End of menu section   ////
  ////------------------------////

  for(kk=0; kk<20; kk++)
  {   aa[kk]   = s.x[mp][kk+1];   // aa params array starts at 0
      dyda[kk] = 1.0;             // set all derivs to 1
  }
  for(row=1; row<=s.maxrow[mx]; row++)
  { 
     xx = s.x[mx][row];
     value = func(funcID, xx, aa, dyda, mfit);
     set_cell_value(row, mz, value); // params are row, col
  }
  if(s.maxrow[mp] > mfit)
  {  sprintf(tempstring, "Only used first %d params in col. %d", mfit, mp);
     message(tempstring);
  }

  setcoltitle(mp, (char*)"Eval.Params");
  setcoltitle(mz, (char*)"Eval.Results");
  refreshspreadsheet();
  delete[] dialog.boxmenutitle[0]; 
  return;
}


